--[[
Precalculation functions

These functions are used to self:get information where clock cycles can be saved by
storing calculated values and only recalculating when the stored value becomes
outdated.
--]]

function AIPlayer:getDistanceToFront(hex,front) -- Find function AIPlayer:of a locus or something? Bounding box
  if front.type == frontType.Sea and front.centerHex ~= nil then
    return math.max(0,front.centerHex:Distance(hex)-1)
  end

  local outleft = false
  local outright = false
  local outtop = false
  local outbottom = false
  
  if hex ~= nil then
  if hex.x < front.boundingBox.left then
    outleft = true
  elseif hex.x > front.boundingBox.right then
    outright = true
  end
  end
  if hex ~= nil then
  if hex.y < front.boundingBox.top then
    outtop = true
  elseif hex.y > front.boundingBox.bottom then
    outbottom = true
  end
  end
  
  if outleft == false and outright == false and outtop == false and outbottom == false then
    return 0
  else
    if outleft then
      if outtop then
        return self:getDistance(hex,front.boundingBox.left,front.boundingBox.top)
      elseif outbottom then
        return self:getDistance(hex,front.boundingBox.left,front.boundingBox.bottom)
      else
        return front.boundingBox.left - hex.x
      end
    elseif outright then
      if outtop then
        return self:getDistance(hex,front.boundingBox.right,front.boundingBox.top)
      elseif outbottom then
        return self:getDistance(hex,front.boundingBox.right,front.boundingBox.bottom)
      else
        return hex.x - front.boundingBox.right
      end
    elseif outtop then
      return front.boundingBox.top - hex.y
    elseif outbottom then
      return hex.y - front.boundingBox.top
    else
      -- error
      self:printLog("AI: ERROR: self:getDistanceToFront defaulting to inf")
      return inf
    end
  end
    -- error
  self:printLog("AI: ERROR: 2 self:getDistanceToFront defaulting to inf")
  return inf
end

function AIPlayer:initPrecalculationTables()
  self.sai.distanceToEnemyCost = {}
  self.sai.nearestUnit = {}
  self.sai.nearestUnit[self.sai.alliance.id]={}
  self.sai.nearestUnit[self.sai.enemyAlliance.id]={}
  self.sai.nearestUnit[self.sai.neutralAlliance.id]={}
  self.sai.nearestNavalUnit = {}
  self.sai.nearestNavalTransport = {}
  self.sai.nearestNavalUnit[self.sai.alliance.id]={}
  self.sai.nearestNavalUnit[self.sai.enemyAlliance.id]={}
  self.sai.nearestNavalUnit[self.sai.neutralAlliance.id]={}
  self.sai.nearestNavalTransport[self.sai.alliance.id]={}
  self.sai.nearestNavalTransport[self.sai.enemyAlliance.id]={}
  self.sai.nearestNavalTransport[self.sai.neutralAlliance.id]={}
  
  self.sai.nearestHostileNeutralUnit = {}
  self.sai.turnMoved = {}
  self.sai.nearestObjective = {}
  self.sai.nearestObjective[self.sai.alliance.id]={}
  self.sai.nearestObjective[self.sai.enemyAlliance.id]={}
  self.sai.nearestObjective[self.sai.neutralAlliance.id]={}
  self.sai.nearestFrontHex = {}
  self.sai.nearestHostileNeutralObjective={}
  self.sai.turnCaptured = {}
  self.sai.supplyCutMoves = {}
end



-- self:getdistanceToEnemyCost: called by findDefensiveFrontline - startTurn --
function AIPlayer:getdistanceToEnemyCost(hex)
  if self.sai.distanceToEnemyCost[hex.id] == nil then
    self.sai.distanceToEnemyCost[hex.id] = {}
    self.sai.distanceToEnemyCost[hex.id].turnUpdated = -1
  end
  
  if self.sai.distanceToEnemyCost[hex.id].turnUpdated < self.sai.cacheTurn then
   dist = self:getDistanceToAlliance(self.sai.enemyAlliance,hex,true)
   self.sai.distanceToEnemyCost[hex.id].cost = self:enemyDistanceCost(dist)
   self.sai.distanceToEnemyCost[hex.id].turnUpdated = self.sai.cacheTurn
  end
  return self.sai.distanceToEnemyCost[hex.id].cost
end
-------------------------------------------------------

function AIPlayer:getNearestNavalUnit(pAlliance,hex) -- Note, hex here is actually a front
  if self.sai.nearestNavalUnit[pAlliance.id][hex.id] == nil or self.sai.nearestNavalUnit[pAlliance.id][hex.id].turnUpdated < self.sai.cacheTurn or self.sai.nearestNavalUnit[pAlliance.id][hex.id].alive ~= true then
    self:updateNearestNavalUnit(pAlliance,hex)
  end
  return self.sai.nearestNavalUnit[pAlliance.id][hex.id].unit, self.sai.nearestNavalUnit[pAlliance.id][hex.id].dist, self.sai.nearestNavalUnit[pAlliance.id][hex.id].dist2
end

function AIPlayer:getNearestNavalTransport(pAlliance,hex) -- Note, hex here is actually a front
  if self.sai.nearestNavalTransport[pAlliance.id][hex.id] == nil or self.sai.nearestNavalTransport[pAlliance.id][hex.id].turnUpdated < self.sai.cacheTurn or self.sai.nearestNavalTransport[pAlliance.id][hex.id].alive ~= true then
    self:updateNearestNavalTransport(pAlliance,hex)
  end
  return self.sai.nearestNavalTransport[pAlliance.id][hex.id].unit, self.sai.nearestNavalTransport[pAlliance.id][hex.id].dist, self.sai.nearestNavalTransport[pAlliance.id][hex.id].dist2
end

-------------------------------------------------------
function AIPlayer:getNearestUnit(pAlliance,hex)
  if self.sai.nearestUnit[pAlliance.id][hex.id] == nil or self.sai.nearestUnit[pAlliance.id][hex.id].turnUpdated < self.sai.cacheTurn or self.sai.nearestUnit[pAlliance.id][hex.id].unit == nil or self.sai.nearestUnit[pAlliance.id][hex.id].unit.alive ~= true then
    self:updateNearestUnit(pAlliance,hex)
  end
  return self.sai.nearestUnit[pAlliance.id][hex.id].unit, self.sai.nearestUnit[pAlliance.id][hex.id].dist, self.sai.nearestUnit[pAlliance.id][hex.id].dist2
end
-------------------------------------------------------
function AIPlayer:getDistanceToHostileNeutral(hex)
  if self.sai.nearestHostileNeutralUnit[hex.id] == nil or self.sai.nearestHostileNeutralUnit[hex.id].turnUpdated < self.sai.cacheTurn or (self.sai.nearestHostileNeutralUnit[hex.id].unit ~= nil and self.sai.nearestHostileNeutralUnit[hex.id].unit.alive ~= true) then
    self:updateNearestHostileNeutralUnit(hex) 
  end
  
  if self.sai.nearestHostileNeutralUnit[hex.id].unit == nil then
    return inf
  end
  
  return self.sai.nearestHostileNeutralUnit[hex.id].dist, self.sai.nearestHostileNeutralUnit[hex.id].dist2, self.sai.nearestHostileNeutralUnit[hex.id].unit, self.sai.nearestHostileNeutralUnit[hex.id].unit2
end

function AIPlayer:getDistanceToAlliance(pAlliance,hex,includeHostileNeutrals)
  if hex == nil then
    self:printLog("AI: ERROR: hex = nil in self:getDistanceToAlliance")
    return inf
  end
  if self.sai.nearestUnit[pAlliance.id][hex.id] == nil or self.sai.nearestUnit[pAlliance.id][hex.id].turnUpdated < self.sai.cacheTurn or (self.sai.nearestUnit[pAlliance.id][hex.id].unit ~= nil and self.sai.nearestUnit[pAlliance.id][hex.id].unit.alive ~= true) then
    self:updateNearestUnit(pAlliance,hex)
  end
  
  local dist1, dist2, dist3, unit1, unit2, unit3
  local ndist1, ndist2, nunit1, nunit2
  if includeHostileNeutrals == true then
    ndist1, ndist2, nunit1, nunit2 = self:getDistanceToHostileNeutral(hex)
  end
  if includeHostileNeutrals == true and ndist1 ~= inf then
    local hdist1, hdist2, hdist3, hunit1, hunit2, hunit3
    hdist1, hdist2, hdist3 = self.sai.nearestUnit[pAlliance.id][hex.id].dist, self.sai.nearestUnit[pAlliance.id][hex.id].dist2, self.sai.nearestUnit[pAlliance.id][hex.id].dist3
    hunit1, hunit2, hunit3 = self.sai.nearestUnit[pAlliance.id][hex.id].unit, self.sai.nearestUnit[pAlliance.id][hex.id].unit2, self.sai.nearestUnit[pAlliance.id][hex.id].unit3
    local all = {{hdist1,hunit1}, {hdist2,hunit2}, {hdist3,hunit3}, {ndist1,nunit1}, {ndist2,nunit2}}
    table.sort(all, function(a,b) 
                      if a[1]==nil then return false end
                      if b[1]==nil then return true end
                      return a[1]<b[1]
                    end)
    dist1 = all[1][1]
    dist2 = all[2][1]
    dist3 = all[3][1]
    unit1 = all[1][2]
    unit2 = all[2][2]
    unit3 = all[3][2]
  else
    dist1, dist2, dist3 = self.sai.nearestUnit[pAlliance.id][hex.id].dist, self.sai.nearestUnit[pAlliance.id][hex.id].dist2, self.sai.nearestUnit[pAlliance.id][hex.id].dist3
    unit1, unit2, unit3 = self.sai.nearestUnit[pAlliance.id][hex.id].unit, self.sai.nearestUnit[pAlliance.id][hex.id].unit2, self.sai.nearestUnit[pAlliance.id][hex.id].unit3
  end
  
  --TODO: Check that second unit hasn't moved and updated if it has?  Also needed for self:get nearest Unit?
  return dist1, dist2, unit1, unit2, dist3, unit3
end
-------------------------------------------------------
function AIPlayer:getDistanceToUnitGroup(units,unitIdAsKeys,hex,pAlliance)
  if hex == nil then
    self:printLog("AI: ERROR: hex = nil in self:getDistanceToUnitGroup")
    return inf
  end
 
  if self.sai.nearestUnit[pAlliance.id][hex.id] == nil or self.sai.nearestUnit[pAlliance.id][hex.id].turnUpdated < self.sai.cacheTurn or self.sai.nearestUnit[pAlliance.id][hex.id].unit.alive ~= true then

  else
    if unitIdAsKeys[self.sai.nearestUnit[pAlliance.id][hex.id].unit.id] ~= false then
      self.sai.nearestUnitInGroup[hex.id] = self.sai.nearestUnit[pAlliance.id][hex.id]
      return self.sai.nearestUnit[pAlliance.id][hex.id].dist
    end
  end

  if self.sai.nearestUnitInGroup[hex.id] ~= nil then
    --printLog("Distance: 6")
    return hex:Distance(self.sai.nearestUnitInGroup[hex.id].hex)
  else
    local bestUnit = nil
    local bestDist = inf
    for _, unit in ipairs(units) do
      if unit.alive and unit.type ~= Unit.NAVAL then
        if unit.hex ~= nil then
          distance = self:getDistance(hex,unit.hex)
          if distance < bestDist then
            bestDist = distance
            bestUnit = unit
          end
        end
      end
    end
    self.sai.nearestUnitInGroup[hex.id] = bestUnit
    return bestDist
  end
  
end

function AIPlayer:isAdjacentToHostileTerritory(hex)
  if self.sai.DI[hex.id] ~= nil and self.sai.DI[hex.id] ~= 0 then
    return true
  end
  local hostileAlliances = self:getHostileFactionIDs()
  table.insert(hostileAlliances,self.sai.enemyAlliance.id)
  local isHostile = {}
  for _,id in ipairs(hostileAlliances) do
    isHostile[id] = true
  end
  if hex.terrain.type ~= Terrain.WATER and isHostile[hex.alliance.id] == true then
    return true
  end
  local neighbours = self:getNeighbours(hex)
  for _,neighbour in pairs(neighbours) do
    if neighbour.terrain.type ~= Terrain.WATER and isHostile[neighbour.alliance.id] == true then
      return true
    end
  end
  return false
end

function AIPlayer:isEnemyLandingSite(hex)
  local neighbours = self:getNeighbours(hex)
  for _,neighbour in pairs(neighbours) do
    if self:unitInHex(neighbour) and neighbour.terrain.type == Terrain.WATER and neighbour.unit.isCarrier then
      return true
    end
  end
  
  return false
end

-----------------------------------------------------
function AIPlayer:isAdjacentToAlliance(pAlliance,hex)
  if self.sai.nearestUnit[pAlliance.id][hex.id] ~= nil then
    if self.sai.nearestUnit[pAlliance.id][hex.id].turnUpdated >= self.sai.cacheTurn then
      if self.sai.nearestUnit[pAlliance.id][hex.id].unit ~= nil and self.sai.nearestUnit[pAlliance.id][hex.id].unit.alive then
        if self.sai.nearestUnit[pAlliance.id][hex.id].dist <= 1 then
          return true
        else
          return false
        end
      end
    end
  end
  
  if self:unitInHex(hex) and hex.unit.type ~= Unit.NAVAL then
    local unitAlliance = hex.unit.alliance
    if unitAlliance.id == pAlliance.id then
      --[[if self.sai.nearestUnit[pAlliance.id][hex.id] == nil then
        self.sai.nearestUnit[pAlliance.id][hex.id] = {}
      end
      self.sai.nearestUnit[pAlliance.id][hex.id].unit = hex.unit
      self.sai.nearestUnit[pAlliance.id][hex.id].turnUpdated = self.sai.cacheTurn
      self.sai.nearestUnit[pAlliance.id][hex.id].dist = 1--]]
      return true
    end
  end

  local neighbours = self:getNeighbours(hex)
  for _,neighbour in pairs(neighbours) do
    if neighbour.unit ~= nil then
      if neighbour.unit.alive and neighbour.unit.type ~= Unit.NAVAL then
        local unitAlliance = neighbour.unit.alliance
        if unitAlliance.id == pAlliance.id then
          --[[if self.sai.nearestUnit[pAlliance.id][hex.id] == nil then
            self.sai.nearestUnit[pAlliance.id][hex.id] = {}
          end
          self.sai.nearestUnit[pAlliance.id][hex.id].unit = neighbour.unit
          self.sai.nearestUnit[pAlliance.id][hex.id].turnUpdated = self.sai.cacheTurn
          self.sai.nearestUnit[pAlliance.id][hex.id].dist = 1--]]
          return true
        end
      end
    end
  end
  
  return false
end
-------------------------------------------------------

-- self:updateNearestHostileNeutralUnit --
function AIPlayer:updateNearestHostileNeutralUnit(hex)

  if self.sai.nearestHostileNeutralUnit[hex.id] == nil then
    self.sai.nearestHostileNeutralUnit[hex.id] = {}
    self.sai.nearestHostileNeutralUnit[hex.id].turnUpdated = -2
    self.sai.nearestHostileNeutralUnit[hex.id].dist = inf
    self.sai.nearestHostileNeutralUnit[hex.id].dist2 = inf
  else
    -- All this could be changed to start by checking if the previous unit had moved or died and use that as initial nearest
    -- however for simplicity start by sorting all.
    -- The below causes a crash because if the unit died more than one turn ago then unit.alive is invalid.
    --[[if self.sai.nearestUnit[pAlliance.id][hex.id].unit.alive then
      self.sai.nearestUnit[pAlliance.id][hex.id].dist = self:getDistance(hex,self.sai.nearestUnit[pAlliance.id][hex.id].unit.hex)
    else--]]
      self.sai.nearestHostileNeutralUnit[hex.id].turnUpdated = -2
      self.sai.nearestHostileNeutralUnit[hex.id].dist = inf
      self.sai.nearestHostileNeutralUnit[hex.id].dist2 = inf
      self.sai.nearestHostileNeutralUnit[hex.id].unit = nil
      self.sai.nearestHostileNeutralUnit[hex.id].unit2 = nil
    --end
  end

  for faction in self.sai.neutralAlliance.factions do

    if self:isFactionHostile(faction) then
      for unit in faction.units do
        if unit.alive and unit.type ~= Unit.NAVAL then
          if unit.hex ~= nil then
            local distance = self:getDistance(hex,unit.hex)
            if distance < self.sai.nearestHostileNeutralUnit[hex.id].dist then
              self.sai.nearestHostileNeutralUnit[hex.id].dist2 = self.sai.nearestHostileNeutralUnit[hex.id].dist
              self.sai.nearestHostileNeutralUnit[hex.id].unit2 = self.sai.nearestHostileNeutralUnit[hex.id].unit
              self.sai.nearestHostileNeutralUnit[hex.id].dist = distance
              self.sai.nearestHostileNeutralUnit[hex.id].unit = unit
            elseif distance < self.sai.nearestHostileNeutralUnit[hex.id].dist2 then
              self.sai.nearestHostileNeutralUnit[hex.id].dist2 = distance
              self.sai.nearestHostileNeutralUnit[hex.id].unit2 = unit
            end
          end
        end
      end
    end

  end
  
  self.sai.nearestHostileNeutralUnit[hex.id].turnUpdated = self.sai.cacheTurn
  
--[[  if self.sai.nearestHostileNeutralUnit[hex.id].unit == nil then
    self:printLog("AI: WARNING no nearest hostile neutral unit!")
  end--]]
end

-- self:updateNearestUnit --
function AIPlayer:updateNearestUnit(pAlliance,hex)

  if self.sai.nearestUnit[pAlliance.id][hex.id] == nil then
    self.sai.nearestUnit[pAlliance.id][hex.id] = {}
    self.sai.nearestUnit[pAlliance.id][hex.id].turnUpdated = -2
    self.sai.nearestUnit[pAlliance.id][hex.id].dist = inf
    self.sai.nearestUnit[pAlliance.id][hex.id].dist2 = inf
    self.sai.nearestUnit[pAlliance.id][hex.id].dist3 = inf
  else
    -- All this could be changed to start by checking if the previous unit had moved or died and use that as initial nearest
    -- however for simplicity start by sorting all.
    -- The below causes a crash because if the unit died more than one turn ago then unit.alive is invalid.
    --[[if self.sai.nearestUnit[pAlliance.id][hex.id].unit.alive then
      self.sai.nearestUnit[pAlliance.id][hex.id].dist = self:getDistance(hex,self.sai.nearestUnit[pAlliance.id][hex.id].unit.hex)
    else--]]
      self.sai.nearestUnit[pAlliance.id][hex.id].turnUpdated = -2
      self.sai.nearestUnit[pAlliance.id][hex.id].dist = inf
      self.sai.nearestUnit[pAlliance.id][hex.id].dist2 = inf
      self.sai.nearestUnit[pAlliance.id][hex.id].dist3 = inf
      self.sai.nearestUnit[pAlliance.id][hex.id].unit = nil
      self.sai.nearestUnit[pAlliance.id][hex.id].unit2 = nil
      self.sai.nearestUnit[pAlliance.id][hex.id].unit3 = nil
    --end
  end
  
  --[[local distance
  if self:unitInHex(hex) and hex.unit.alliance.id == pAlliance.id then
    distance = 0
    self.sai.nearestUnit[pAlliance.id][hex.id].dist = 0
    self.sai.nearestUnit[pAlliance.id][hex.id].unit = hex.unit
  end--]] -- Only do this if only one unit is neaded, then break
  
  -- TODO add parameter to flag if two units are needed.  If not, skip the next step & break if distance == 0
  if pAlliance.units == nil then
    return
  end
  
  for _, unit in ipairs(pAlliance.units) do
    if unit.alive and unit.type ~= Unit.NAVAL then
      if unit.hex ~= nil then
        local distance = self:getDistance(hex,unit.hex)
        if distance <= self.sai.nearestUnit[pAlliance.id][hex.id].dist then
          self.sai.nearestUnit[pAlliance.id][hex.id].dist3 = self.sai.nearestUnit[pAlliance.id][hex.id].dist2
          self.sai.nearestUnit[pAlliance.id][hex.id].unit3 = self.sai.nearestUnit[pAlliance.id][hex.id].unit2
          
          self.sai.nearestUnit[pAlliance.id][hex.id].dist2 = self.sai.nearestUnit[pAlliance.id][hex.id].dist
          self.sai.nearestUnit[pAlliance.id][hex.id].unit2 = self.sai.nearestUnit[pAlliance.id][hex.id].unit
          
          self.sai.nearestUnit[pAlliance.id][hex.id].dist = distance
          self.sai.nearestUnit[pAlliance.id][hex.id].unit = unit
          
        elseif distance < self.sai.nearestUnit[pAlliance.id][hex.id].dist2 then
          self.sai.nearestUnit[pAlliance.id][hex.id].dist3 = self.sai.nearestUnit[pAlliance.id][hex.id].dist2
          self.sai.nearestUnit[pAlliance.id][hex.id].unit3 = self.sai.nearestUnit[pAlliance.id][hex.id].unit2

          self.sai.nearestUnit[pAlliance.id][hex.id].dist2 = distance
          self.sai.nearestUnit[pAlliance.id][hex.id].unit2 = unit
          
        elseif distance < self.sai.nearestUnit[pAlliance.id][hex.id].dist3 then
          self.sai.nearestUnit[pAlliance.id][hex.id].dist3 = distance
          self.sai.nearestUnit[pAlliance.id][hex.id].unit3 = unit
          if distance == 1 then
            -- The third nearest unit is adjacent
            break
          end
        end
      end
    end
  end
  
  self.sai.nearestUnit[pAlliance.id][hex.id].turnUpdated = self.sai.cacheTurn
  
  if self.sai.nearestUnit[pAlliance.id][hex.id].unit == nil then
    self:printLog("AI: WARNING no nearest unit!")
  end
end
-------------------------------------------------------

-- self:updateNearestUnit --
function AIPlayer:updateNearestNavalUnit(pAlliance,hex) -- NOTE Hex here is actually a front -- NOt any more

  if self.sai.nearestNavalUnit[pAlliance.id][hex.id] == nil then
    self.sai.nearestNavalUnit[pAlliance.id][hex.id] = {}
    self.sai.nearestNavalUnit[pAlliance.id][hex.id].turnUpdated = -2
    self.sai.nearestNavalUnit[pAlliance.id][hex.id].dist = inf
    self.sai.nearestNavalUnit[pAlliance.id][hex.id].dist2 = inf
    self.sai.nearestNavalUnit[pAlliance.id][hex.id].dist3 = inf
  else
      self.sai.nearestNavalUnit[pAlliance.id][hex.id].turnUpdated = -2
      self.sai.nearestNavalUnit[pAlliance.id][hex.id].dist = inf
      self.sai.nearestNavalUnit[pAlliance.id][hex.id].dist2 = inf
      self.sai.nearestNavalUnit[pAlliance.id][hex.id].dist3 = inf
      self.sai.nearestNavalUnit[pAlliance.id][hex.id].unit = nil
      self.sai.nearestNavalUnit[pAlliance.id][hex.id].unit2 = nil
      self.sai.nearestNavalUnit[pAlliance.id][hex.id].unit3 = nil
  end

  for _, unit in ipairs(pAlliance.units) do
    if unit.alive and unit.type == Unit.NAVAL and unit.carrier ~= true then
      if unit.hex ~= nil then
        local distance = self:getDistance(unit.hex,hex)-- self:getDistanceToFront(unit.hex,hex)-- NOTE Hex here is actually a front
        if distance <= self.sai.nearestNavalUnit[pAlliance.id][hex.id].dist then
          self.sai.nearestNavalUnit[pAlliance.id][hex.id].dist3 = self.sai.nearestNavalUnit[pAlliance.id][hex.id].dist2
          self.sai.nearestNavalUnit[pAlliance.id][hex.id].unit3 = self.sai.nearestNavalUnit[pAlliance.id][hex.id].unit2
          
          self.sai.nearestNavalUnit[pAlliance.id][hex.id].dist2 = self.sai.nearestNavalUnit[pAlliance.id][hex.id].dist
          self.sai.nearestNavalUnit[pAlliance.id][hex.id].unit2 = self.sai.nearestNavalUnit[pAlliance.id][hex.id].unit
          
          self.sai.nearestNavalUnit[pAlliance.id][hex.id].dist = distance
          self.sai.nearestNavalUnit[pAlliance.id][hex.id].unit = unit
          
        elseif distance < self.sai.nearestNavalUnit[pAlliance.id][hex.id].dist2 then
          self.sai.nearestNavalUnit[pAlliance.id][hex.id].dist3 = self.sai.nearestNavalUnit[pAlliance.id][hex.id].dist2
          self.sai.nearestNavalUnit[pAlliance.id][hex.id].unit3 = self.sai.nearestNavalUnit[pAlliance.id][hex.id].unit2

          self.sai.nearestNavalUnit[pAlliance.id][hex.id].dist2 = distance
          self.sai.nearestNavalUnit[pAlliance.id][hex.id].unit2 = unit
          
        elseif distance < self.sai.nearestNavalUnit[pAlliance.id][hex.id].dist3 then
          self.sai.nearestNavalUnit[pAlliance.id][hex.id].dist3 = distance
          self.sai.nearestNavalUnit[pAlliance.id][hex.id].unit3 = unit
          if distance == 1 then
            -- The third nearest unit is adjacent
            break
          end
        end
      end
    end
  end
  
  self.sai.nearestNavalUnit[pAlliance.id][hex.id].turnUpdated = self.sai.cacheTurn
  
  if self.sai.nearestNavalUnit[pAlliance.id][hex.id].unit == nil then
    self:printLog("AI: WARNING no nearest naval unit!")
  end
end
-------------------------------------------------------


function AIPlayer:updateNearestNavalTransport(pAlliance,hex) -- NOTE Hex here is actually a front -- NOt any more

  if self.sai.nearestNavalTransport[pAlliance.id][hex.id] == nil then
    self.sai.nearestNavalTransport[pAlliance.id][hex.id] = {}
    self.sai.nearestNavalTransport[pAlliance.id][hex.id].turnUpdated = -2
    self.sai.nearestNavalTransport[pAlliance.id][hex.id].dist = inf
  end

  for _, unit in ipairs(pAlliance.units) do
    if unit.alive and unit.type == Unit.NAVAL and unit.isCarrier then
      if unit.hex ~= nil then
        local distance = self:getDistance(unit.hex,hex)-- self:getDistanceToFront(unit.hex,hex)-- NOTE Hex here is actually a front
        if distance <= self.sai.nearestNavalTransport[pAlliance.id][hex.id].dist then
          self.sai.nearestNavalTransport[pAlliance.id][hex.id].dist = distance
          self.sai.nearestNavalTransport[pAlliance.id][hex.id].unit = unit
        end
      end
    end
  end
  
  self.sai.nearestNavalTransport[pAlliance.id][hex.id].turnUpdated = self.sai.cacheTurn
  
  if self.sai.nearestNavalTransport[pAlliance.id][hex.id].unit == nil then
    self:printLog("AI: WARNING no nearest naval transport unit!")
  end
end


function AIPlayer:getDistanceToFrontline(front,hex)
  if #front.frontline == 0 then
    return 0
  end
  if self.sai.nearestFrontHex[front.id] == nil then
    self.sai.nearestFrontHex[front.id] = {}
  end
  if self.sai.nearestFrontHex[front.id][hex.id] == nil then
    self.sai.nearestFrontHex[front.id][hex.id] = {}
    self.sai.nearestFrontHex[front.id][hex.id].turnUpdated = -2
    self.sai.nearestFrontHex[front.id][hex.id].dist = inf
  end
  
  if self.sai.nearestFrontHex[front.id][hex.id].turnUpdated < front.frontlineCacheCounter then
    self:updateNearestFrontHex(front,hex)
  end
  return self.sai.nearestFrontHex[front.id][hex.id].dist
end

function AIPlayer:updateNearestFrontHex(front,hex)
  if self.sai.nearestFrontHex[front.id] == nil then
    self.sai.nearestFrontHex[front.id] = {}
  end
  if self.sai.nearestFrontHex[front.id][hex.id] == nil then
    self.sai.nearestFrontHex[front.id][hex.id] = {}
    self.sai.nearestFrontHex[front.id][hex.id].turnUpdated = -2
    self.sai.nearestFrontHex[front.id][hex.id].dist = inf
  end

  for _,frontHex in pairs(front.frontline) do
    local distance = self:getDistance(hex,frontHex)
    if distance < self.sai.nearestFrontHex[front.id][hex.id].dist then
      self.sai.nearestFrontHex[front.id][hex.id].dist = distance
      self.sai.nearestFrontHex[front.id][hex.id].frontHex = frontHex
    end
    if distance == 0 then
      break
    end
  end
  
  self.sai.nearestFrontHex[front.id][hex.id].turnUpdated = front.frontlineCacheCounter
end

-------------------------------------------------------
function AIPlayer:getNearestObjective(pAlliance,hex)
  if self.sai.nearestObjective[pAlliance.id][hex.id] == nil then
    self.sai.nearestObjective[pAlliance.id][hex.id] = {}
    self.sai.nearestObjective[pAlliance.id][hex.id].turnUpdated = -2
    self.sai.nearestObjective[pAlliance.id][hex.id].dist = inf
    self.sai.nearestObjective[pAlliance.id][hex.id].dist2 = inf
  end
  
  if self.sai.nearestObjective[pAlliance.id][hex.id].turnUpdated < game.turn or self.sai.nearestObjective[pAlliance.id][hex.id].objective.owner.id ~= self.sai.nearestObjective[pAlliance.id][hex.id].objective.hex.alliance.id then
    self:updateNearestObjective(pAlliance,hex)
  end
  return self.sai.nearestObjective[pAlliance.id][hex.id].objective, self.sai.nearestObjective[pAlliance.id][hex.id].dist, self.sai.nearestObjective[pAlliance.id][hex.id].objective2, self.sai.nearestObjective[pAlliance.id][hex.id].dist2
end
-------------------------------------------------------
function AIPlayer:getDistanceToObjective(pAlliance,hex,includeHostileNeutrals)
  if self.sai.nearestObjective[pAlliance.id][hex.id] == nil then
    self.sai.nearestObjective[pAlliance.id][hex.id] = {}
    self.sai.nearestObjective[pAlliance.id][hex.id].turnUpdated = -2
    self.sai.nearestObjective[pAlliance.id][hex.id].dist = inf
    self.sai.nearestObjective[pAlliance.id][hex.id].dist2 = inf
  end

  if self.sai.nearestObjective[pAlliance.id][hex.id].turnUpdated < game.turn or self.sai.nearestObjective[pAlliance.id][hex.id].objective.owner.id ~= self.sai.nearestObjective[pAlliance.id][hex.id].objective.hex.alliance.id then
    self:updateNearestObjective(pAlliance,hex)
  end
  
  local obj = self.sai.nearestObjective[pAlliance.id][hex.id].objective
  local dist1 = self.sai.nearestObjective[pAlliance.id][hex.id].dist
  local obj2 = self.sai.nearestObjective[pAlliance.id][hex.id].objective2
  local dist2 = self.sai.nearestObjective[pAlliance.id][hex.id].dist2
  if includeHostileNeutrals == true then
    local ndist, nobj
    ndist, nobj = self:getDistanceToHostileNeutralObjective(hex)
    if ndist < dist1 then
      dist2 = dist1
      obj2 = obj
      dist1 = ndist
      obj = nobj
    elseif ndist < dist2 then
      dist2 = ndist
      obj2 = nobj
    end
  end
  
  return dist1, obj, dist2, obj
end
-------------------------------------------------------

-------------------------------------------------------
function AIPlayer:getDistanceToHostileNeutralObjective(hex)
  if self.sai.nearestHostileNeutralObjective[hex.id] == nil then
    self.sai.nearestHostileNeutralObjective[hex.id] = {}
    self.sai.nearestHostileNeutralObjective[hex.id].turnUpdated = -2
    self.sai.nearestHostileNeutralObjective[hex.id].dist = inf
  end

  if self.sai.nearestHostileNeutralObjective[hex.id].turnUpdated < game.turn then
    self:updateNearestHostileNeutralObjective(hex)
  end
  return self.sai.nearestHostileNeutralObjective[hex.id].dist, self.sai.nearestHostileNeutralObjective[hex.id].objective
end
-------------------------------------------------------

-- self:updateNearestHostileNeutralObjective --
function AIPlayer:updateNearestHostileNeutralObjective(hex)
  for _,objective in ipairs(self.sai.neutralAlliance.objectives) do
    -- disabled as in function AIPlayer:below
    --[[if self.sai.turnCaptured[objective] == nil then
      self.sai.turnCaptured[objective] = game.turn
    end--]]

    --if self.sai.turnCaptured[objective] >= self.sai.nearestObjective[pAlliance.id][hex.id].turnUpdated then
      --distance = hex:Distance(objective.hex)
    if self:isFactionHostile(objective.hex.faction) then
      local distance = self:getDistance(hex,objective.hex)
      if distance < self.sai.nearestHostileNeutralObjective[hex.id].dist then
        self.sai.nearestHostileNeutralObjective[hex.id].dist = distance
        self.sai.nearestHostileNeutralObjective[hex.id].objective = objective
      end
    end
    --end
  end

  self.sai.nearestHostileNeutralObjective[hex.id].turnUpdated = game.turn
end
-------------------------------------------------------

-- self:updateNearestObjective --
function AIPlayer:updateNearestObjective(pAlliance,hex)
  for _,objective in ipairs(pAlliance.objectives) do
    -- Don't think this is working, disable for not
    --[[if self.sai.turnCaptured[objective] == nil then
      self.sai.turnCaptured[objective] = game.turn
    end--]]

    --if self.sai.turnCaptured[objective] >= self.sai.nearestObjective[pAlliance.id][hex.id].turnUpdated then
      --distance = hex:Distance(objective.hex)
      local distance = self:getDistance(hex,objective.hex)
      local capitalBonus = 0
      if objective.hex.construction.type == Construction.TYPE_CAPITAL then -- Where capital and city are equidistant
        capitalBonus = 0.5
      end
      if distance - capitalBonus < self.sai.nearestObjective[pAlliance.id][hex.id].dist then
        objective.owner = objective.hex.alliance
        self.sai.nearestObjective[pAlliance.id][hex.id].dist2 = self.sai.nearestObjective[pAlliance.id][hex.id].dist
        self.sai.nearestObjective[pAlliance.id][hex.id].objective2 = self.sai.nearestObjective[pAlliance.id][hex.id].objective
        self.sai.nearestObjective[pAlliance.id][hex.id].dist = distance
        self.sai.nearestObjective[pAlliance.id][hex.id].objective = objective
      elseif distance - capitalBonus < self.sai.nearestObjective[pAlliance.id][hex.id].dist2 then
        self.sai.nearestObjective[pAlliance.id][hex.id].dist2 = distance
        self.sai.nearestObjective[pAlliance.id][hex.id].objective2 = objective
      end
    --end
  end

  self.sai.nearestObjective[pAlliance.id][hex.id].turnUpdated = game.turn
end
-------------------------------------------------------

-- self:updateNearestObjective --
function AIPlayer:getNearestPort(pAlliance,hex)
  local bestDist = inf
  local bestHex
  for _,objective in ipairs(pAlliance.objectives) do
    if objective.hex.construction ~= nil and objective.hex.construction.hasPort then
      if self:HexRegionsLinked(objective.hex.construction.port.hex,hex) then
        local distance = self:getDistance(hex,objective.hex)
        if distance < bestDist then
          bestDist = distance
          bestHex = objective.hex
        end
      end
    end
  end

  return bestHex
end
-------------------------------------------------------

function AIPlayer:getNearestFriendlyCoast(hex)
  local bestDist = inf
  local bestHex
  for _,area in pairs(self.sai.areas) do
    for _,coast in ipairs(area.coast) do
      local coastSeaHex
      _,coastSeaHex = self:isCoastalHex(coast)
      if self:HexRegionsLinked(coastSeaHex, hex) then
        local distance = self:getDistance(hex,coast)
        if distance < bestDist then
          bestDist = distance
          bestHex = coast
        end
      end
    end
  end
  return bestHex
end
  
-- findObjectives
function AIPlayer:findObjectives()
  local objectives = {}
  local capitals = {}
  --for _,pAlliance in ipairs(alliances) do
    for faction in game.factions do
      if faction.constructions ~= nil then
        for construction in faction.constructions do
          local hex = construction.hex
          if hex.terrain ~= Terrain.WATER and hex.faction ~= nil then -- TODO: remove reference to hex.faction when working properly
            local objective = {}
            objective.hex = hex
            objective.value = 50
            objective.id = objective.hex.id
            objective.name = construction.name
            if construction.type==2 then
              objective.value = 100
              table.insert(capitals,objective)
            end
            table.insert(objectives,objective)
          end
        end
      end
    end  
  --end

  return objectives, capitals
end

-- updateObjectiveOwnership: called by findDefensiveFrontline - startTurn --
function AIPlayer:updateObjectiveOwnership()
  --printLog("Updating objective ownership")
  self.sai.alliance.objectives = {}
  self.sai.enemyAlliance.objectives = {}
  self.sai.getCapitals = {}
  self.sai.getCapitals[self.sai.alliance.id] = {}
  self.sai.getCapitals[self.sai.enemyAlliance.id] = {}
  self.sai.getCapitals[self.sai.neutralAlliance.id] = {}
  local changedObjectives = {}
  local hexAlliance
  local capitalsChanged = false

  for _,capital in ipairs(self.sai.capitals) do
    if capital.owner == nil then
      capital.owner = capital.hex.alliance
      capitalsChanged = true
    end  

    hexAlliance = capital.hex.alliance
    if hexAlliance.id ~= capital.owner.id then
      capital.owner = hexAlliance
      capitalsChanged = true
    end 
    
    table.insert(self.sai.getCapitals[capital.owner.id],capital)
  end
  
  if capitalsChanged == true then
    for _,objective in ipairs(self.sai.objectives) do
      objective.importance = 30
      ---[[
      for _,capital in ipairs(self.sai.getCapitals[self.sai.alliance.id]) do
        --printLog("Distance: 5")
        local dist = capital.hex:Distance(objective.hex)
        dist = 100 - dist*2
        if dist > objective.importance then
            objective.importance = dist
        end
      end--]]
    end
  end

  for _,objective in ipairs(self.sai.objectives) do
      if objective.owner == nil then
        objective.owner = objective.hex.alliance
      end

      hexAlliance = objective.hex.alliance
      if hexAlliance.id ~= objective.owner.id then
        objective.owner = hexAlliance
        table.insert(changedObjectives,objective)
      end
      --printLog("AI: Alliance = " .. hexAlliance.name .. " Faction = " .. objective.hex.faction.name)

      if objective.hex.terrain.type == Terrain.LAND then
        if hexAlliance.id == self.sai.alliance.id then
          table.insert(self.sai.alliance.objectives,objective)
        elseif hexAlliance.id == self.sai.enemyAlliance.id then
          table.insert(self.sai.enemyAlliance.objectives,objective)
        elseif hexAlliance.id == self.sai.neutralAlliance.id then
          table.insert(self.sai.neutralAlliance.objectives,objective)
        end
      end
      
    --end
  end

  
  return changedObjectives
end
-------------------------------------------------------


function AIPlayer:updateAllianceLists()
  -- update alliance unit and hex lists
  self.sai.alliance.units = {}
  self.sai.artilleryUnits = {}
  self.sai.alliance.hexes = {}
  self.sai.alliance.fronts = {}
  self.sai.alliance.coast = {}
  for faction in self.sai.alliance.factions do
    for unit in faction.units do
      if unit.alive then
        table.insert(self.sai.alliance.units,unit)
        
        if unit.prototype.lua.class ~= "artillery" then
          if self.sai.artilleryUnits[unit.faction.id] == nil then
            self.sai.artilleryUnits[unit.faction.id] = {}
          end
          table.insert(self.sai.artilleryUnits[unit.faction.id],unit)
        end
      end
    end
    for hex in faction.hexes do
      table.insert(self.sai.alliance.hexes,hex)
      if self:isCoastalHex(hex) then
        table.insert(self.sai.alliance.coast,hex)
      end
    end
  end

  self.sai.alliance.undeployedUnits = self:getUndeployedUnits(self.sai.alliance)
  
  
  self.sai.enemyAlliance.units = {}
  self.sai.enemyAlliance.hexes = {}
  for faction in self.sai.enemyAlliance.factions do
    for unit in faction.units do
      if unit.alive then
        table.insert(self.sai.enemyAlliance.units,unit)
      end
    end
    for hex in faction.hexes do
      table.insert(self.sai.enemyAlliance.hexes,hex)
    end
  end
end